@@ -38,6 +38,7 @@
 #include "shrpx_http.h"
 #include "shrpx_http2_session.h"
 #include "shrpx_worker.h"
+#include "shrpx_log.h"
 #include "http2.h"
 #include "util.h"
 
@@ -202,7 +203,7 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
       nva.reserve(trailers.size());
       // We cannot use nocopy version, since nva may be touched after
       // Downstream object is deleted.
-      http2::copy_headers_to_nva(nva, trailers);
+      http2::copy_headers_to_nva(nva, trailers, http2::HDOP_STRIP_ALL);
       if (!nva.empty()) {
         rv = nghttp2_submit_trailer(session, stream_id, nva.data(), nva.size());
         if (rv != 0) {
@@ -246,11 +247,11 @@ int Http2DownstreamConnection::push_request_headers() {
 
   auto &balloc = downstream_->get_block_allocator();
 
-  auto &httpconf = get_config()->http;
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &httpconf = config->http;
+  auto &http2conf = config->http2;
 
-  auto no_host_rewrite = httpconf.no_host_rewrite ||
-                         get_config()->http2_proxy ||
+  auto no_host_rewrite = httpconf.no_host_rewrite || config->http2_proxy ||
                          req.method == HTTP_CONNECT;
 
   // http2session_ has already in CONNECTED state, so we can get
@@ -309,7 +310,16 @@ int Http2DownstreamConnection::push_request_headers() {
     nva.push_back(http2::make_nv_ls_nocopy(":authority", authority));
   }
 
-  http2::copy_headers_to_nva_nocopy(nva, req.fs.headers());
+  auto &fwdconf = httpconf.forwarded;
+  auto &xffconf = httpconf.xff;
+  auto &xfpconf = httpconf.xfp;
+
+  uint32_t build_flags =
+      (fwdconf.strip_incoming ? http2::HDOP_STRIP_FORWARDED : 0) |
+      (xffconf.strip_incoming ? http2::HDOP_STRIP_X_FORWARDED_FOR : 0) |
+      (xfpconf.strip_incoming ? http2::HDOP_STRIP_X_FORWARDED_PROTO : 0);
+
+  http2::copy_headers_to_nva_nocopy(nva, req.fs.headers(), build_flags);
 
   if (!http2conf.no_cookie_crumbling) {
     downstream_->crumble_request_cookie(nva);
@@ -318,15 +328,13 @@ int Http2DownstreamConnection::push_request_headers() {
   auto upstream = downstream_->get_upstream();
   auto handler = upstream->get_client_handler();
 
-  auto &fwdconf = httpconf.forwarded;
-
   auto fwd =
       fwdconf.strip_incoming ? nullptr : req.fs.header(http2::HD_FORWARDED);
 
   if (fwdconf.params) {
     auto params = fwdconf.params;
 
-    if (get_config()->http2_proxy || req.method == HTTP_CONNECT) {
+    if (config->http2_proxy || req.method == HTTP_CONNECT) {
       params &= ~FORWARDED_PROTO;
     }
 
@@ -350,14 +358,12 @@ int Http2DownstreamConnection::push_request_headers() {
     nva.push_back(http2::make_nv_ls_nocopy("forwarded", fwd->value));
   }
 
-  auto &xffconf = httpconf.xff;
-
   auto xff = xffconf.strip_incoming ? nullptr
                                     : req.fs.header(http2::HD_X_FORWARDED_FOR);
 
   if (xffconf.add) {
     StringRef xff_value;
-    auto addr = StringRef{upstream->get_client_handler()->get_ipaddr()};
+    const auto &addr = upstream->get_client_handler()->get_ipaddr();
     if (xff) {
       xff_value = concat_string_ref(balloc, xff->value,
                                     StringRef::from_lit(", "), addr);
@@ -369,9 +375,24 @@ int Http2DownstreamConnection::push_request_headers() {
     nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-for", xff->value));
   }
 
-  if (!get_config()->http2_proxy && req.method != HTTP_CONNECT) {
-    // We use same protocol with :scheme header field
-    nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", req.scheme));
+  if (!config->http2_proxy && req.method != HTTP_CONNECT) {
+    auto xfp = xfpconf.strip_incoming
+                   ? nullptr
+                   : req.fs.header(http2::HD_X_FORWARDED_PROTO);
+
+    if (xfpconf.add) {
+      StringRef xfp_value;
+      // We use same protocol with :scheme header field
+      if (xfp) {
+        xfp_value = concat_string_ref(balloc, xfp->value,
+                                      StringRef::from_lit(", "), req.scheme);
+      } else {
+        xfp_value = req.scheme;
+      }
+      nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", xfp_value));
+    } else if (xfp) {
+      nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", xfp->value));
+    }
   }
 
   auto via = req.fs.header(http2::HD_VIA);
@@ -402,7 +423,7 @@ int Http2DownstreamConnection::push_request_headers() {
   // HTTP/1 upstream request can contain keyword other than
   // "trailers".  We just forward "trailers".
   // TODO more strict handling required here.
-  if (te && util::strifind(te->value, StringRef::from_lit("trailers"))) {
+  if (te && http2::contains_trailers(te->value)) {
     nva.push_back(http2::make_nv_ll("te", "trailers"));
   }
 
@@ -421,22 +442,27 @@ int Http2DownstreamConnection::push_request_headers() {
 
   auto transfer_encoding = req.fs.header(http2::HD_TRANSFER_ENCODING);
 
+  nghttp2_data_provider *data_prdptr = nullptr;
+  nghttp2_data_provider data_prd;
+
   // Add body as long as transfer-encoding is given even if
   // req.fs.content_length == 0 to forward trailer fields.
   if (req.method == HTTP_CONNECT || transfer_encoding ||
       req.fs.content_length > 0 || req.http2_expect_body) {
     // Request-body is expected.
-    nghttp2_data_provider data_prd{{}, http2_data_read_callback};
-    rv = http2session_->submit_request(this, nva.data(), nva.size(), &data_prd);
-  } else {
-    rv = http2session_->submit_request(this, nva.data(), nva.size(), nullptr);
+    data_prd = {{}, http2_data_read_callback};
+    data_prdptr = &data_prd;
   }
+
+  rv = http2session_->submit_request(this, nva.data(), nva.size(), data_prdptr);
   if (rv != 0) {
     DCLOG(FATAL, this) << "nghttp2_submit_request() failed";
     return -1;
   }
 
-  downstream_->reset_downstream_wtimer();
+  if (data_prdptr) {
+    downstream_->reset_downstream_wtimer();
+  }
 
   http2session_->signal_write();
   return 0;
